/*
 * Copyright  2016 TIBCO Software Inc. All Rights Reserved.
 * This software is confidential and proprietary information of TIBCO Software Inc.
 */
package com.streambase.contrib.batch;

import java.net.URL;

import com.streambase.sb.Schema;
import com.streambase.sb.StreamBaseException;
import com.streambase.sb.Tuple;
import com.streambase.sb.operator.Operator;
import com.streambase.sb.operator.Parameterizable;
import com.streambase.sb.operator.TypecheckException;

/**
 * Batch operator entry point. The batch operator annotates incoming tuples
 * with batch id and sequence number within batch. Once batch is detected
 * as closed, an asynchronous tuple is emitted with batch id, grouping key
 * and effective batch size.
 * 
 * @author <a href="mailto:pismolin@tibco.com">Piotr Smolinski</a>
 */
public class BatchOperator extends Operator implements Parameterizable {

	public static final long serialVersionUID = 1455610746235L;

	// ---- properties ---------------------------------------------------------

	private long batchSize;
	private long timeout;
	private String[] groupByFields;

	// ---- runtime state ------------------------------------------------------

	/**
	 * Operator runtime reference
	 */
	private BatchOperatorRuntime runtime = null;

	// ---- basic contract -----------------------------------------------------

	public BatchOperator() {

		super();

		setPortHints(1, 2);
		setDisplayName("Batch");
		setShortDisplayName(this.getClass().getSimpleName());

		setBatchSize(0);
		setTimeout(0L);
		setGroupByFields(new String[] {});

	}

	@Override
	public URL getIconResource(IconKind iconType) {
		switch (iconType) {
			case CANVAS_OVERLAY_ICON:
				return getClass().getResource("icons/batch-48.png");
			case PALETTE_ICON_LARGE:
				return getClass().getResource("icons/batch-32.png");
			case PALETTE_ICON_SMALL:
				return getClass().getResource("icons/batch-16.png");
		}
		return null;
	}

	/**
	 * Design time configuration validation. The schema verification and generation
	 * is delegated to {@link BatchOperatorSchemas}.
	 */
	@Override
	public void typecheck() throws TypecheckException {
		requireInputPortCount(1);
		if (batchSize<=0) {
			throw new PropertyTypecheckException("batchSize", "Batch size must be greater than zero");
		}
		if (timeout<=0) {
			throw new PropertyTypecheckException("timeout", "Timeout must be greater than zero");
		}
		Schema inputSchema = getInputSchema(0);
		BatchOperatorSchemas.typecheckInputSchema(this, inputSchema);
		BatchOperatorSchemas.typecheckGroupByFields(this, inputSchema, groupByFields);
		setOutputSchema(0, BatchOperatorSchemas.generateOutputSchema(inputSchema, groupByFields));
		setOutputSchema(1, BatchOperatorSchemas.generateBatchSchema(inputSchema, groupByFields));
	}

	// ---- runtime lifecycle --------------------------------------------------

	/**
	 * Instantiate and configure the runtime. Note that suspending and resuming
	 * is not supported. 
	 */
	public void init() throws StreamBaseException {
		super.init();
		Schema inputSchema = getRuntimeInputSchema(0);
		BatchOperatorRuntime runtime = new BatchOperatorRuntime();
		runtime.setBatchSize(batchSize);
		runtime.setTimeout(timeout);
		runtime.setGroupByFields(groupByFields);
		runtime.setOperator(this);
		runtime.setInputSchema(inputSchema);
		runtime.setOutputSchema(BatchOperatorSchemas.generateOutputSchema(inputSchema, groupByFields));
		runtime.setBatchSchema(BatchOperatorSchemas.generateBatchSchema(inputSchema, groupByFields));
		runtime.init();
		this.runtime = runtime;
	}
	
	/**
	 * Delegate the call to the runtime
	 */
	public void processTuple(int inputPort, Tuple tuple) throws StreamBaseException {
		runtime.processTuple(tuple);
	}

	/**
	 * Terminate the runtime
	 */
	public void shutdown() {
		if (this.runtime==null) return;
		try {
			runtime.shutdown();
		} finally {
			this.runtime = null;
		}
	}

	// ---- operator properties accessors --------------------------------------

	public void setTimeout(long timeout) {
		this.timeout = timeout;
	}

	public long getTimeout() {
		return this.timeout;
	}

	public void setBatchSize(long batchSize) {
		this.batchSize = batchSize;
	}
	
	public long getBatchSize() {
		return batchSize;
	}

	public void setGroupByFields(String[] groupByFields) {
		this.groupByFields = groupByFields;
	}

	public String[] getGroupByFields() {
		return this.groupByFields;
	}

}
