/*
 * Copyright  2016 TIBCO Software Inc. All Rights Reserved.
 * This software is confidential and proprietary information of TIBCO Software Inc.
 */
package com.streambase.contrib.batch;

import java.io.Serializable;
import java.util.Arrays;
import java.util.Deque;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

import com.streambase.sb.Schema;
import com.streambase.sb.StreamBaseException;
import com.streambase.sb.Tuple;
import com.streambase.sb.TupleCopier;
import com.streambase.sb.operator.Operator;

/**
 * Runtime implementation of the batch operator. The class decouples the runtime
 * behaviour from the design time and contract component
 * @author <a href="mailto:pismolin@tibco.com">Piotr Smolinski</a>
 *
 */
public class BatchOperatorRuntime implements Serializable, Runnable {

	private static final long serialVersionUID = 1455610746235L;

	// 	---- dependencies -------------------------------------------------------

	/**
	 * Parent operator/environment reference
	 */
	private Operator operator;

	// ---- configuration ------------------------------------------------------
	
	private long batchSize;
	private long timeout;
	private String[] groupByFields;

	// ---- cached schemas -----------------------------------------------------

	private Schema inputSchema;
	private Schema outputSchema;
	private TupleCopier outputCopier;
	private Schema batchSchema;

	// ---- runtime state ------------------------------------------------------
	
	private Map<Fields, Batch> currentBatches = new HashMap<>(); 
	private Map<Fields, Long> lastBatchId = new HashMap<>(); 
	private Deque<Batch> emissionQueue = new LinkedList<Batch>();

	private Thread timeoutHandler;
	
	// ---- coordination -------------------------------------------------------
	
	private ReentrantLock lock = new ReentrantLock(false);
	private Condition notEmpty = lock.newCondition();
	private Condition trigger = lock.newCondition();

	// ---- public contract ----------------------------------------------------

	/**
	 * For every incoming tuple assign a batchId and batchSeq. If given batch 
	 * exceeds its size, emit it.
	 * @param tuple
	 * @throws StreamBaseException
	 */
	public void processTuple(Tuple tuple) throws StreamBaseException {

		Object[] g = new Object[groupByFields.length];
		for (int i=0;i<groupByFields.length;i++) {
			g[i] = tuple.getField(groupByFields[i]);
		}
		Fields fg = new Fields(g);
		long emissionTime = System.currentTimeMillis()+timeout;

		lock.lock();
		try {

			Batch batch = currentBatches.get(fg);
			if (batch==null) {
				Long batchId = lastBatchId.get(fg);
				if (batchId==null) {
					batchId = 0L;
				} else {
					batchId = batchId+1;
				}
				lastBatchId.put(fg, batchId);
				batch = new Batch(fg, batchId, emissionTime);
				currentBatches.put(fg, batch);
				emissionQueue.add(batch);
				notEmpty.signalAll();
			}

			long batchSeq = batch.allocateBatchSeq();
			Tuple outputTuple = outputSchema.createTuple();
			outputCopier.copyTuple(tuple, outputTuple);
			outputTuple.setLong("batchId", batch.getBatchId());
			outputTuple.setLong("batchSeq", batchSeq);
			operator.sendOutput(0, outputTuple);

			if (batch.getBatchCount()>=batchSize) {
				batch.setEmitted();
				currentBatches.remove(fg, batch);
				Tuple batchTuple = batchSchema.createTuple();
				for (int i=0;i<groupByFields.length;i++) {
					batchTuple.setField(groupByFields[i], fg.getFields()[i]);
				}
				batchTuple.setLong("batchId", batch.getBatchId());
				batchTuple.setLong("batchCount", batch.getBatchCount());
				operator.sendOutputAsync(1, batchTuple);
			}

		} finally {
			lock.unlock();
		}

	}
	
	/**
	 * Timeout handling thread. The thread waits until there are any batches.
	 * Then it waits till emission time and, if the batch has not been emitted
	 * yet, it emits the batch tuple.
	 */
	public void run() {
		lock.lock();
		try {
			while (!Thread.currentThread().isInterrupted()) {
				if (emissionQueue.isEmpty()) {
					notEmpty.await();
					continue;
				}
				Batch batch = emissionQueue.peek();
				long t = batch.emissionTime-System.currentTimeMillis();
				if (t>0) {
					trigger.await(t, TimeUnit.MILLISECONDS);
				}
				batch = emissionQueue.poll();
				if (batch.isEmitted()) {
					// batch may have been emitted based on size
					continue;
				}
				Fields fg = batch.getGroupByFields();
				batch.setEmitted();
				currentBatches.remove(fg, batch);
				try {
					Tuple batchTuple = batchSchema.createTuple();
					for (int i=0;i<groupByFields.length;i++) {
						batchTuple.setField(groupByFields[i], fg.getFields()[i]);
					}
					batchTuple.setLong("batchId", batch.getBatchId());
					batchTuple.setLong("batchCount", batch.getBatchCount());
					operator.sendOutputAsync(1, batchTuple);
				} catch (StreamBaseException e) {
					operator.sendErrorOutput(e);
				}
			}
		} catch (InterruptedException e) {
			// just terminate
		} finally {
			lock.unlock();
		}
	}

	/**
	 * Initializes tuple copier and timeout handler
	 */
	public void init() throws StreamBaseException {
		this.outputCopier = new TupleCopier(inputSchema, outputSchema, false);
		if (this.timeout>0) {
			this.timeoutHandler = new Thread(this, operator.getFullyQualifiedName()+".TimeoutHandler");
			this.timeoutHandler.start();
		}
	}

	/**
	 * Terminates the timeout handler.
	 */
	public void shutdown() {
		if (this.timeoutHandler==null) return;
		this.timeoutHandler.interrupt();
		try {
			this.timeoutHandler.join();
		} catch (InterruptedException e) {
			// restore interrupted status and let the caller handle it
			Thread.currentThread().interrupt();
		} finally {
			this.timeoutHandler = null;
		}
	}

	// ---- dependency injection -----------------------------------------------

	public void setOperator(Operator operator) {
		this.operator = operator;
	}

	public void setBatchSize(long batchSize) {
		this.batchSize = batchSize;
	}

	public void setTimeout(long timeout) {
		this.timeout = timeout;
	}

	public void setGroupByFields(String[] groupByFields) {
		this.groupByFields = groupByFields;
	}
	
	public void setBatchSchema(Schema batchSchema) {
		this.batchSchema = batchSchema;
	}
	
	public void setOutputSchema(Schema outputSchema) {
		this.outputSchema = outputSchema;
	}
	
	public void setInputSchema(Schema inputSchema) {
		this.inputSchema = inputSchema;
	}

	// ---- utility classes ----------------------------------------------------

	/**
	 * Batch information:
	 */
	private static class Batch {

		private final Fields groupByFields;
		private final long emissionTime;
		private final long batchId;

		private long batchSeq = 0;

		private boolean emitted;

		public Batch(Fields groupByFields, long batchId, long emissionTime) {
			this.groupByFields = groupByFields;
			this.batchId = batchId;
			this.emissionTime = emissionTime;
		}
		
		public Fields getGroupByFields() {
			return groupByFields;
		}

		public long getBatchId() {
			return batchId;
		}
		
		public long allocateBatchSeq() {
			return batchSeq++;
		}

		public long getBatchCount() {
			return batchSeq;
		}

		public boolean isEmitted() {
			return emitted;
		}

		public void setEmitted() {
			this.emitted = true;
		}

	}

	/**
	 * Comparable wrapper around fields.
	 */
	private static class Fields implements Comparable<Fields>{

		private final Object[] fields;
		public Fields(Object[] fields) {
			this.fields = fields;
		}

		@SuppressWarnings({ "rawtypes", "unchecked" })
		@Override
		public int compareTo(Fields that) {
			for (int i=0;i<this.fields.length && i<that.fields.length;i++) {
				Object thisField = this.fields[i]; 
				Object thatField = that.fields[i]; 
				if (thisField==null && thatField==null) continue;
				if (thisField==null) return -1;
				if (thatField==null) return +1;
				Comparable thisC = (Comparable) thisField;
				Comparable thatC = (Comparable) thatField;
				int c = thisC.compareTo(thatC);
				if (c!=0) return c;
			}
			return 0;
		}
		
		@Override
		public int hashCode() {
			return Arrays.hashCode(fields);
		}
		
		@Override
		public boolean equals(Object o) {
			if (!(o instanceof Fields)) return false;
			Fields that = (Fields) o;
			return Arrays.equals(this.fields, that.fields);
		}
		
		public Object[] getFields() {
			return fields;
		}

	}

}
