/*
 * Copyright  2016 TIBCO Software Inc. All Rights Reserved.
 * This software is confidential and proprietary information of TIBCO Software Inc.
 */
package com.streambase.contrib.batch;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import com.streambase.sb.DataType;
import com.streambase.sb.Schema;
import com.streambase.sb.TupleException;
import com.streambase.sb.operator.Operator;
import com.streambase.sb.operator.TypecheckException;

/**
 * Schema handling logic for {@link BatchOperator}.
 * @author <a href="mailto:pismolin@tibco.com">Piotr Smolinski</a>
 */
public class BatchOperatorSchemas {
	
	/**
	 * Verify if the inputSchema does not declare fields used by the operator
	 * @param operator reference to the operator
	 * @param inputSchema verified input schema
	 */
	public static void typecheckInputSchema(Operator operator, Schema inputSchema) throws TypecheckException {
		if (inputSchema.hasField("batchId")) {
			throw new TypecheckException("Unexpected field 'batchId'");
		}
		if (inputSchema.hasField("batchSeq")) {
			throw new TypecheckException("Unexpected field 'batchSeq'");
		}
	}

	/**
	 * Verify if the declared group by fields are present in the input schema
	 * and if they have simple types. 
	 * @param operator reference to the operator
	 * @param inputSchema verified input schema
	 * @param groupByFields list of the fields (may be empty)
	 */
	public static void typecheckGroupByFields(Operator operator, Schema inputSchema, String[] groupByFields) throws TypecheckException {
		Set<String> checkedFields = new HashSet<String>();
		for (String fieldName : groupByFields) {
			if ("batchCount".equals(fieldName)) {
				throw operator.new PropertyTypecheckException("groupByFields", "Unexpected group by field 'batchCount'");
			}
			if (checkedFields.contains(fieldName)) {
				throw operator.new PropertyTypecheckException("groupByFields", "Duplicate group by field '"+fieldName+"'");
			}
			checkedFields.add(fieldName);
			Schema.Field field = getField(inputSchema, fieldName);
			if (field==null) {
				throw operator.new PropertyTypecheckException("groupByFields", "Field '"+fieldName+"' does not exist in the input schema");
			}
			switch (field.getDataType()) {
			case BLOB:
			case CAPTURE:
			case TUPLE:
			case LIST:
			case FUNCTION:
				throw operator.new PropertyTypecheckException("groupByFields", "Field '"+fieldName+"' must have primitive type");
			default:
			}
		}
	}
	
	/**
	 * Generates the base output schema by adding the batchId and batchSeq fields
	 */
	public static Schema generateOutputSchema(Schema inputSchema, String[] groupByFields) {
		List<Schema.Field> fields = new ArrayList<Schema.Field>(inputSchema.getFieldCount()+2);
		fields.addAll(Arrays.asList(inputSchema.getFields()));
		fields.add(Schema.createField(DataType.LONG, "batchId"));
		fields.add(Schema.createField(DataType.LONG, "batchSeq"));
		return new Schema(null, fields);
	}

	/**
	 * Generates the batch notification schema by selecting the grouping fields
	 * and adding the batchId and batchCount fields
	 */
	public static Schema generateBatchSchema(Schema inputSchema, String[] groupByFields) {
		List<Schema.Field> fields = new ArrayList<Schema.Field>(groupByFields.length+2);
		for (int i=0;i<groupByFields.length;i++) {
			fields.add(getField(inputSchema, groupByFields[i]));
		}
		fields.add(Schema.createField(DataType.LONG, "batchId"));
		fields.add(Schema.createField(DataType.LONG, "batchCount"));
		return new Schema(null, fields);
	}

	/**
	 * Utility method for hiding the {@link TupleException} that is never thrown.
	 */
	private static Schema.Field getField(Schema schema, String fieldName) {
		if (!schema.hasField(fieldName)) return null;
		try {
			return schema.getField(fieldName);
		} catch (TupleException e) {
			// unexpected
			return null;
		}
	}

}
